/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: EnhancedPool.java 4961 2009-05-17 11:44:57Z gaellalire $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.pool.impl.enhanced;

import java.util.concurrent.Executor;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.api.IPoolConfiguration;
import org.ow2.util.pool.api.PoolException;
import org.ow2.util.pool.impl.enhanced.api.IPool;
import org.ow2.util.pool.impl.enhanced.api.TimeoutPoolException;
import org.ow2.util.pool.impl.enhanced.api.keepbusy.ShareMethod;
import org.ow2.util.pool.impl.enhanced.api.resizer.IResizablePool;
import org.ow2.util.pool.impl.enhanced.impl.basic.BasicPool;
import org.ow2.util.pool.impl.enhanced.impl.basic.accessmanager.LastUsedAccessManager;
import org.ow2.util.pool.impl.enhanced.impl.keepbusy.KeepBusyPool;
import org.ow2.util.pool.impl.enhanced.impl.limited.LimitedWaiterPool;
import org.ow2.util.pool.impl.enhanced.impl.limited.TooManyWaiterException;
import org.ow2.util.pool.impl.enhanced.impl.listener.resizer.StatsResizerPool;
import org.ow2.util.pool.impl.enhanced.impl.listener.statistics.BasicPoolStats;
import org.ow2.util.pool.impl.enhanced.impl.listener.statistics.BasicPoolStatsFillerPoolListener;
import org.ow2.util.pool.impl.enhanced.impl.util.ExecutorProvider;
import org.ow2.util.pool.impl.enhanced.impl.validator.ValidatorPool;
import org.ow2.util.pool.impl.enhanced.internal.resizer.api.IResizer;
import org.ow2.util.pool.impl.enhanced.internal.resizer.api.shared.ISharedManager;
import org.ow2.util.pool.impl.enhanced.internal.resizer.impl.AsynchronousResizer;
import org.ow2.util.pool.impl.enhanced.internal.resizer.impl.shared.SharedResizer;
import org.ow2.util.pool.impl.enhanced.internal.resizer.impl.stocker.MinMaxExpectedSpareStockerResizer;
import org.ow2.util.pool.impl.enhanced.manager.IPoolManager;
import org.ow2.util.pool.impl.enhanced.manager.optional.IValidatorPoolManager;

/**
 *
 * @author Gael Lalire
 * @param <E>
 */
public class EnhancedPool<E> {

    private static final int DEFAULT_SPARE_POOL = 150;

    private static final Log LOG = LogFactory.getLog(EnhancedPool.class);

    private IPool<E> lastPool;

    private LimitedWaiterPool<E> limitedWaiterPool;

    private KeepBusyPool<E> keepBusyPool;

    private MinMaxExpectedSpareStockerResizer<BasicPoolStats> minMaxExpectedSparePoolResizer;

    private long timeout;

    private int maxPool = 0;

    private int maxWaiter = LimitedWaiterPool.NO_LIMIT_WAITER;

    private boolean started = true;

    private Object mutex = new Object();

    public EnhancedPool(final IPoolManager<E> poolManager) {
        this(poolManager, ResizerType.SIMPLE, null, ExecutorProvider.NEW_THREAD_EXECUTOR);
    }

    public EnhancedPool(final IPoolManager<E> poolManager, final ResizerType resizerType, final ISharedManager sharedManager, final Executor createItemExecutor) {
        LimitedWaiterPool<E> limitedWaiterPool;
        KeepBusyPool<E> keepBusyPool;
        StatsResizerPool<E, BasicPoolStats> statsResizerPool;
        ValidatorPool<E> validatorPool;
        IResizablePool<E> pool;

        IPool<E> currentCluePool;
        // 1 clue pool
        pool = new BasicPool<E>(poolManager, 0, new LastUsedAccessManager<E>(),
                createItemExecutor);

        minMaxExpectedSparePoolResizer = new MinMaxExpectedSpareStockerResizer<BasicPoolStats>(0, 0, DEFAULT_SPARE_POOL, true);

        IResizer<BasicPoolStats> resizer = minMaxExpectedSparePoolResizer;

        switch (resizerType) {
        case ASYNCHRONOUS:
            resizer = new AsynchronousResizer<BasicPoolStats>(minMaxExpectedSparePoolResizer);
            break;
        default:
        case SHARED_ASYNCHRONOUS:
            resizer = new SharedResizer<BasicPoolStats>(sharedManager, minMaxExpectedSparePoolResizer);
            break;
        case SIMPLE:
            break;
        }
        // IResizer<BasicPoolStats> resizer = minMaxExpectedSparePoolResizer;
        resizer.setResizable(pool);
        currentCluePool = pool;

        // 2 validator
        if (IValidatorPoolManager.class.isInstance(poolManager)) {
            IValidatorPoolManager<E> validatorCluePoolManager = (IValidatorPoolManager<E>) poolManager;
            validatorPool = new ValidatorPool<E>(currentCluePool, validatorCluePoolManager, false);
            currentCluePool = validatorPool;
        }

        // 3.0 stats & resize pool
        statsResizerPool = new StatsResizerPool<E, BasicPoolStats>(currentCluePool,
                new BasicPoolStatsFillerPoolListener<E>(), new BasicPoolStats(), resizer);
        // 3.1 add a 3e listener
        // FIXME manage remove listener and activate / passivate listener => create new optional managers
        // statsResizerCluePool.addCluePoolListener(poolFactoryImpl);
        currentCluePool = statsResizerPool;

        // 4 keep busy pool
        keepBusyPool = new KeepBusyPool<E>(currentCluePool);
        // 5 limited waiter pool
        limitedWaiterPool = new LimitedWaiterPool<E>(keepBusyPool);
        currentCluePool = limitedWaiterPool;
        // end pool construction

        this.limitedWaiterPool = limitedWaiterPool;
        this.keepBusyPool = keepBusyPool;
        this.lastPool = currentCluePool;

        timeout = 0;
        setAllowSharedInstance(false);
        setPoolConfiguration(new PoolConfiguration());
    }

    public E get() throws PoolException {
        try {
            LOG.debug("START GET {1} try to get with timeout {0}", timeout);
            E e = lastPool.get(timeout);
            LOG.debug("OK FOR GET {1} try to get with timeout {0}", timeout);
            return e;
        } catch (TimeoutPoolException e) {
            throw new PoolException("No more instances available", e);
        } catch (TooManyWaiterException e) {
            throw new PoolException("No more instances available", e);
        } catch (org.ow2.util.pool.impl.enhanced.api.PoolException e) {
            throw new PoolException("Unknow exception", e);
        }
    }

    public void discard(final E instance) throws PoolException {
        LOG.debug("discard {0} ", instance);
        try {
            lastPool.remove(instance);
        } catch (org.ow2.util.pool.impl.enhanced.api.PoolException e) {
            throw new PoolException("Unknow exception", e);
        }
    }

    public void release(final E instance) throws PoolException {
        LOG.debug("release {0} ", instance);
        try {
            lastPool.put(instance);
        } catch (org.ow2.util.pool.impl.enhanced.api.PoolException e) {
            throw new PoolException("Unknow exception", e);
        }
    }

    public void setPoolConfiguration(final IPoolConfiguration poolConfiguration) {
        maxPool = poolConfiguration.getMax();
        timeout = poolConfiguration.getTimeout();
        maxWaiter = poolConfiguration.getMaxWaiters();
        limitedWaiterPool.setMaxWaiter(maxWaiter);
        minMaxExpectedSparePoolResizer.setMaxPool(maxPool);
        minMaxExpectedSparePoolResizer.update();
        LOG.debug("{3} new conf Setted (timeout : {0}, maxwaiter: {1}, max: {2})", timeout, maxWaiter,
                maxPool, this);
    }

    /**
     * Allow to share instances (shouldn't be use in stateful case as one
     * stateful ID is linked to one client. No Stateful with same ID are
     * authorized).
     * @param allowSharedInstance true if it is allowed, else false.
     */
    public void setAllowSharedInstance(final boolean allowSharedInstance) {
        if (allowSharedInstance) {
            keepBusyPool.setShareMethod(ShareMethod.SHARE_AS_LAST_SOLUTION);
        } else {
            keepBusyPool.setShareMethod(ShareMethod.NEVER_SHARE);
        }
    }

    /**
     * Return true if many instances of the same object can be created (when
     * pool is not full).
     * @return true if this is allowed.
     */
    public boolean isAllowSharedInstance() {
        return keepBusyPool.getShareMethod() != ShareMethod.NEVER_SHARE;
    }

    public void start() throws PoolException {
        synchronized (mutex) {
            if (started) {
                throw new PoolException("Pool already started");
            }
            limitedWaiterPool.setMaxWaiter(maxWaiter);
            minMaxExpectedSparePoolResizer.setMaxPool(maxPool);
            started = true;
        }
    }

    public void stop() throws PoolException {
        synchronized (mutex) {
            if (!started) {
                throw new PoolException("Pool already stopped");
            }
            // new call to get will result in a TooManyWaiterException.
            limitedWaiterPool.setMaxWaiter(0);
            // interrupt current waiters
            lastPool.interruptAllWaiters();
            // remove busy instances
            keepBusyPool.removeAll();
            // remove not busy instances
            minMaxExpectedSparePoolResizer.setMaxPool(0);
            minMaxExpectedSparePoolResizer.update();
            // now all pool items must be removed from pool
            started = false;
        }
    }

}
